@@ -1,27 +1,56 @@
-
-import base64
-import sys
 import re
 import logging
 from webtest.forms import Upload
 import webtest
 import mimetypes
-import mock
+
 import pytest
 import py
+from bs4 import BeautifulSoup
+from devpi_server.config import PluginManager
 from devpi_server.main import XOM, parseoptions
 from devpi_common.url import URL
 from devpi_server.extpypi import XMLProxy
-import hashlib
+from devpi_server.extpypi import PyPIStage
+from devpi_server.log import threadlog, thread_clear_log
+from pyramid.authentication import b64encode
 
-log = logging.getLogger(__name__)
+import hashlib
+try:
+    from queue import Queue as BaseQueue
+except ImportError:
+    from Queue import Queue as BaseQueue
+
+class TimeoutQueue(BaseQueue):
+    def get(self, timeout=2):
+        return BaseQueue.get(self, timeout=timeout)
+
+log = threadlog
 
 def pytest_addoption(parser):
-    parser.addoption("--catchall", action="store_true", default=False,
-        help="run bottle apps in catchall mode to see exceptions")
     parser.addoption("--slow", action="store_true", default=False,
         help="run slow tests involving remote services (pypi.python.org)")
 
+@pytest.fixture(autouse=True)
+def _clear():
+    thread_clear_log()
+
+
+@pytest.yield_fixture
+def pool():
+    from devpi_server.mythread  import ThreadPool
+    pool = ThreadPool()
+    yield pool
+    pool.shutdown()
+
+@pytest.fixture
+def queue():
+    return TimeoutQueue()
+
+@pytest.fixture
+def Queue():
+    return TimeoutQueue
+
 @pytest.fixture()
 def caplog(caplog):
     """ enrich the pytest-capturelog funcarg. """
@@ -59,34 +88,94 @@ def gentmp(request):
 def xom_notmocked(request, makexom):
     return makexom(mocking=False)
 
-@pytest.fixture
-def xom(request, makexom):
-    return makexom([])
+@pytest.yield_fixture(autouse=True)
+def auto_transact(request):
+    names = request.fixturenames
+    if ("xom" not in names and "keyfs" not in names) or (
+        request.node.get_marker("notransaction")):
+        yield
+        return
+    keyfs = request.getfuncargvalue("keyfs")
+
+    write = True if request.node.get_marker("writetransaction") else False
+    keyfs.begin_transaction_in_thread(write=write)
+    yield
+    try:
+        keyfs.rollback_transaction_in_thread()
+    except AttributeError:  # already finished within the test
+        pass
+
 
 @pytest.fixture
-def makexom(request, gentmp, httpget):
-    def makexom(opts=(), httpget=httpget, proxy=None, mocking=True):
+def xom(request, makexom):
+    xom = makexom([])
+    return xom
+
+@pytest.yield_fixture(autouse=True, scope="session")
+def speed_up_sql():
+    from devpi_server.keyfs import Filesystem
+    old = Filesystem.get_sqlconn
+    def make_unsynchronous(self, old=old):
+        conn = old(self)
+        conn.execute("PRAGMA synchronous=OFF")
+        return conn
+    Filesystem.get_sqlconn = make_unsynchronous
+    yield
+    Filesystem.get_sqlconn = old
+
+@pytest.fixture(scope="session")
+def mock():
+    try:
+        from unittest import mock
+    except ImportError:
+        import mock
+    return mock
+
+@pytest.fixture
+def makexom(request, gentmp, httpget, monkeypatch, mock):
+    def makexom(opts=(), httpget=httpget, proxy=None, mocking=True, plugins=()):
+        hook = PluginManager(plugins)
         serverdir = gentmp()
         fullopts = ["devpi-server", "--serverdir", serverdir] + list(opts)
         fullopts = [str(x) for x in fullopts]
-        config = parseoptions(fullopts)
+        config = parseoptions(fullopts, hook=hook)
         if mocking:
             if proxy is None:
                 proxy = mock.create_autospec(XMLProxy)
                 proxy.list_packages_with_serial.return_value = {}
             xom = XOM(config, proxy=proxy, httpget=httpget)
-            add_extdb_mocks(xom.extdb, httpget)
+            add_pypistage_mocks(monkeypatch, httpget)
         else:
             xom = XOM(config)
-        request.addfinalizer(xom.shutdown)
+        # initialize default indexes
+        from devpi_server.main import set_default_indexes
+        if not xom.config.args.master_url:
+            with xom.keyfs.transaction(write=True):
+                set_default_indexes(xom.model)
+        if mocking:
+            xom.pypimirror.init_pypi_mirror(proxy)
+        if request.node.get_marker("start_threads"):
+            xom.thread_pool.start()
+        elif request.node.get_marker("with_notifier"):
+            xom.thread_pool.start_one(xom.keyfs.notifier)
+        request.addfinalizer(xom.thread_pool.shutdown)
         return xom
     return makexom
 
 
+@pytest.fixture
+def replica_xom(request, makexom):
+    from devpi_server.replica import PyPIProxy
+    master_url = "http://localhost:3111"
+    xom = makexom(["--master", master_url])
+    xom.proxy = PyPIProxy(xom._httpsession, xom.config.master_url)
+    return xom
+
+
 @pytest.fixture
 def maketestapp(request):
     def maketestapp(xom):
-        app = xom.create_app(catchall=False, immediatetasks=-1)
+        app = xom.create_app()
         mt = MyTestApp(app)
         mt.xom = xom
         return mt
@@ -132,7 +221,7 @@ def httpget(pypiurls):
             log.debug("set mocking response %s %s", mockurl, kw)
             self.url2response[mockurl] = kw
 
-        def setextsimple(self, name, text=None, pkgver=None,
+        def mock_simple(self, name, text=None, pkgver=None,
             pypiserial=10000, **kw):
             class ret:
                 md5 = None
@@ -177,36 +266,37 @@ def keyfs(xom):
     return xom.keyfs
 
 @pytest.fixture
-def extdb(xom):
-    return xom.extdb
+def model(xom):
+    return xom.model
 
-def add_extdb_mocks(extdb, httpget):
+@pytest.fixture
+def pypistage(xom):
+    return PyPIStage(xom)
+
+def add_pypistage_mocks(monkeypatch, httpget):
     # add some mocking helpers
-    extdb.url2response = httpget.url2response
-    def setextsimple(name, text=None, pypiserial=10000, **kw):
-        extdb._set_project_serial(name, pypiserial)
-        return httpget.setextsimple(name,
+    PyPIStage.url2response = httpget.url2response
+    def mock_simple(self, name, text=None, pypiserial=10000, **kw):
+        call = lambda: \
+                 self.pypimirror.process_changelog([(name, 0,0,0, pypiserial)])
+        if not hasattr(self.keyfs, "tx"):
+            with self.keyfs.transaction(write=True):
+                call()
+        else:
+            call()
+        return self.httpget.mock_simple(name,
                 text=text, pypiserial=pypiserial, **kw)
-    extdb.setextsimple = setextsimple
-    extdb.mock_simple = setextsimple
-    extdb.httpget = httpget
+    monkeypatch.setattr(PyPIStage, "mock_simple", mock_simple, raising=False)
 
 @pytest.fixture
 def pypiurls():
-    from devpi_server.extpypi import PYPIURL_SIMPLE, PYPIURL
+    from devpi_server.extpypi import PYPIURL, PYPIURL_SIMPLE
     class PyPIURL:
         def __init__(self):
             self.base = PYPIURL
             self.simple = PYPIURL_SIMPLE
     return PyPIURL()
 
-@pytest.fixture
-def db(xom):
-    from devpi_server.db import DB
-    from devpi_server.main import set_default_indexes
-    db = DB(xom)
-    set_default_indexes(db)
-    return db
 
 @pytest.fixture
 def mapp(makemapp, testapp):
@@ -227,6 +317,10 @@ class Mapp(MappMixin):
             return self.current_stage
         return indexname
 
+    def _wait_for_serial_in_result(self, r):
+        commit_serial = int(r.headers["X-DEVPI-SERIAL"])
+        self.xom.keyfs.notifier.wait_event_serial(commit_serial)
+
     def delete_user(self, user, code=200):
         r = self.testapp.delete_json("/%s" % user, expect_errors=True)
         assert r.status_code == code
@@ -246,16 +340,20 @@ class Mapp(MappMixin):
         self.login("root", "")
 
     def getuserlist(self):
-        r = self.testapp.get("/", {"indexes": False}, {"Accept": "*/json"})
+        r = self.testapp.get("/", {"indexes": False}, {"Accept": "application/json"})
         assert r.status_code == 200
         return r.json["result"]
 
     def getindexlist(self, user=None):
         if user is None:
             user = self.testapp.auth[0]
-        r = self.testapp.get("/%s/" % user, {"Accept": "*/json"})
+        r = self.testapp.get("/%s" % user, {"Accept": "*/json"})
         assert r.status_code == 200
-        return r.json["result"]
+        name = r.json["result"]["username"]
+        result = {}
+        for index, data in r.json["result"].get("indexes", {}).items():
+            result["%s/%s" % (name, index)] = data
+        return result
 
     def change_password(self, user, password):
         r = self.testapp.patch_json("/%s" % user, dict(password=password))
@@ -329,7 +427,22 @@ class Mapp(MappMixin):
         if code in (400,):
             return r.json["message"]
 
-    def delete_index(self, indexname, code=201):
+    def modify_index(self, indexname, indexconfig, code=200):
+        if "/" in indexname:
+            user, index = indexname.split("/")
+        else:
+            user, password = self.testapp.auth
+            index = indexname
+        r = self.testapp.patch_json("/%s/%s" % (user, index), indexconfig,
+                                  expect_errors=True)
+        assert r.status_code == code
+        if code in (200,201):
+            assert r.json["result"]["type"] == "stage"
+            return r.json["result"]
+        if code in (400,):
+            return r.json["message"]
+
+    def delete_index(self, indexname, code=201, waithooks=False):
         if "/" in indexname:
             user, index = indexname.split("/")
         else:
@@ -337,8 +450,19 @@ class Mapp(MappMixin):
             index = indexname
         r = self.testapp.delete_json("/%s/%s" % (user, index),
                                      expect_errors=True)
+        if waithooks:
+            self._wait_for_serial_in_result(r)
         assert r.status_code == code
 
+    def set_custom_data(self, data, indexname=None):
+        indexname = self._getindexname(indexname)
+        indexurl = "/" + indexname
+        r = self.testapp.get_json(indexurl)
+        result = r.json["result"]
+        result["custom_data"] = data
+        r = self.testapp.patch_json(indexurl, result)
+        assert r.status_code == 200
+
     def set_uploadtrigger_jenkins(self, triggerurl, indexname=None):
         indexname = self._getindexname(indexname)
         indexurl = "/" + indexname
@@ -348,6 +472,17 @@ class Mapp(MappMixin):
         r = self.testapp.patch_json(indexurl, result)
         assert r.status_code == 200
 
+    def set_pypi_whitelist(self, whitelist, indexname=None):
+        indexname = self._getindexname(indexname)
+        r = self.testapp.get_json("/%s" % indexname)
+        result = r.json["result"]
+        if not isinstance(whitelist, list):
+            whitelist = whitelist.split(",")
+        assert isinstance(whitelist, list)
+        result["pypi_whitelist"] = whitelist
+        r = self.testapp.patch_json("/%s" % (indexname,), result)
+        assert r.status_code == 200
+
     def set_acl(self, users, acltype="upload", indexname=None):
         indexname = self._getindexname(indexname)
         r = self.testapp.get_json("/%s" % indexname)
@@ -364,24 +499,41 @@ class Mapp(MappMixin):
         r = self.testapp.get_json("/%s" % indexname)
         return r.json["result"].get("acl_" + acltype, None)
 
-    def delete_project(self, projectname, code=200, indexname=None):
+    def get_pypi_whitelist(self, indexname=None):
+        indexname = self._getindexname(indexname)
+        r = self.testapp.get_json("/%s" % indexname)
+        return r.json["result"]["pypi_whitelist"]
+
+    def delete_project(self, projectname, code=200, indexname=None,
+                       waithooks=False):
         indexname = self._getindexname(indexname)
         r = self.testapp.delete_json("/%s/%s" % (indexname,
                 projectname), {}, expect_errors=True)
         assert r.status_code == code
+        if waithooks:
+            self._wait_for_serial_in_result(r)
 
-    def register_metadata(self, metadata, indexname=None, code=200):
+    def set_versiondata(self, metadata, indexname=None, code=200,
+                          waithooks=False,
+                          set_whitelist=True):
         indexname = self._getindexname(indexname)
         metadata = metadata.copy()
         metadata[":action"] = "submit"
         r = self.testapp.post("/%s/" % indexname, metadata,
                               expect_errors=True)
         assert r.status_code == code
+        if r.status_code == 200 and set_whitelist:
+            whitelist = set(self.get_pypi_whitelist(indexname=indexname))
+            whitelist.add(metadata["name"])
+            self.set_pypi_whitelist(sorted(whitelist), indexname=indexname)
+        if waithooks:
+            self._wait_for_serial_in_result(r)
+        return r
 
     def upload_file_pypi(self, basename, content,
                          name=None, version=None, indexname=None,
-                         register=True,
-                         code=200):
+                         register=True, code=200, waithooks=False,
+                         set_whitelist=True):
         assert py.builtin._isbytes(content)
         indexname = self._getindexname(indexname)
         #name_version = splitbasename(basename, checkarch=False)
@@ -390,29 +542,63 @@ class Mapp(MappMixin):
         #if not version:
         #    version = name_version[1]
         if register and code == 200:
-            self.register_metadata(dict(name=name, version=version))
+            self.set_versiondata(
+                dict(name=name, version=version), set_whitelist=set_whitelist)
         r = self.testapp.post("/%s/" % indexname,
             {":action": "file_upload", "name": name, "version": version,
              "content": Upload(basename, content)}, expect_errors=True)
         assert r.status_code == code
+        if waithooks:
+            self._wait_for_serial_in_result(r)
+        return r
 
+    def get_release_paths(self, projectname):
+        r = self.get_simple(projectname)
+        pkg_url = URL(r.request.url)
+        paths = [pkg_url.joinpath(link["href"]).path
+                 for link in BeautifulSoup(r.body).findAll("a")]
+        return paths
 
     def upload_doc(self, basename, content, name, version, indexname=None,
-                         code=200):
+                         code=200, waithooks=False):
         indexname = self._getindexname(indexname)
-        r = self.testapp.post("/%s/" % indexname,
-            {":action": "doc_upload", "name": name, "version": version,
-             "content": Upload(basename, content)}, expect_errors=True)
+        form = {":action": "doc_upload", "name": name,
+                "content": Upload(basename, content)}
+        if version:
+            form["version"] = version
+        r = self.testapp.post("/%s/" % indexname, form, expect_errors=True)
         assert r.status_code == code
+        if waithooks:
+            self._wait_for_serial_in_result(r)
+        return r
 
     def get_simple(self, projectname, code=200):
-        r = self.testapp.get(self.api.simpleindex + projectname + "/",
+        r = self.testapp.get(self.api.simpleindex + projectname,
                              expect_errors=True)
         assert r.status_code == code
         return r
 
 
 from webtest import TestApp as TApp
+from webtest import TestResponse
+
+
+@pytest.yield_fixture
+def noiter(monkeypatch, request):
+    l = []
+    @property
+    def body(self):
+        if self.headers["Content-Type"] != "application/octet-stream":
+            return self.body_old
+        if self.app_iter:
+            l.append(self.app_iter)
+    monkeypatch.setattr(TestResponse, "body_old", TestResponse.body,
+                        raising=False)
+    monkeypatch.setattr(TestResponse, "body", body)
+    yield
+    for x in l:
+        x.close()
+
 
 class MyTestApp(TApp):
     auth = None
@@ -425,13 +611,7 @@ class MyTestApp(TApp):
             headers = kw.get("headers")
             if not headers:
                 headers = kw["headers"] = {}
-            auth = "%s:%s" % self.auth
-            if sys.version_info[0] >= 3:
-                res = "Basic " + base64.b64encode(auth.encode("ascii")).decode("ascii")
-            else:
-                res = "Basic %s" % base64.b64encode(auth)
-            #base64.b64encode(("%s:%s" % self.auth).encode("ascii"))
-            headers["Authorization"] = res
+            headers["X-Devpi-Auth"] = b64encode("%s:%s" % self.auth)
             #print ("setting auth header %r %s %s" % (auth, method, url))
         return super(MyTestApp, self)._gen_request(method, url, **kw)
 
@@ -444,15 +624,21 @@ class MyTestApp(TApp):
             assert r.status_code == code
         return r
 
+
     def push(self, url, params=None, **kw):
         kw.setdefault("expect_errors", True)
-        return self._gen_request("push", url, params=params, **kw)
+        return self._gen_request("PUSH", url, params=params, **kw)
 
     def get(self, *args, **kwargs):
-        if "expect_errors" not in kwargs:
-            kwargs["expect_errors"] = True
+        kwargs.setdefault("expect_errors", True)
         return super(MyTestApp, self).get(*args, **kwargs)
 
+    def xget(self, code, *args, **kwargs):
+        r = self.get(*args, **kwargs)
+        assert r.status_code == code
+        return r
+
+
     def get_json(self, *args, **kwargs):
         headers = kwargs.setdefault("headers", {})
         headers["Accept"] = "application/json"
@@ -488,6 +674,7 @@ from requests.adapters import HTTPAdapter
 from requests.packages.urllib3.response import HTTPResponse
 import fnmatch
 
+
 @pytest.fixture
 def reqmock(monkeypatch):
     mr = mocked_request()
@@ -535,6 +722,7 @@ class mocked_request:
             method = method.upper()
         self.url2reply[(url, method)] = r
         return r
+    mock = mockresponse
 
 class ReqReply(HTTPResponse):
     def __init__(self, code, data, headers, on_request):
@@ -573,3 +761,19 @@ class Gen:
 
 def getmd5(s):
     return hashlib.md5(s.encode("utf8")).hexdigest()
+
+
+@pytest.yield_fixture
+def dummyrequest():
+    from pyramid.testing import DummyRequest, setUp, tearDown
+    request = DummyRequest()
+    setUp(request=request)
+    yield request
+    tearDown()
+
+@pytest.fixture
+def blank_request():
+    from pyramid.request import Request
+    def blank_request(*args, **kwargs):
+        return Request.blank("/blankpath", *args, **kwargs)
+    return blank_request
