@@ -289,10 +289,13 @@ class TestSubmitValidation:
     def test_upload_file(self, submit):
         metadata = {"name": "Pkg5", "version": "1.0", ":action": "submit"}
         submit.metadata(metadata, code=200)
-        submit.file("pkg5-2.6.tgz", b"123", {"name": "pkg5some"}, code=400)
+        r = submit.file("pkg5-2.6.tgz", b"123", {"name": "pkg5some"}, code=400)
+        assert "no project" in r.status
         submit.file("pkg5-2.6.tgz", b"123", {"name": "Pkg5"}, code=200)
-        submit.file("pkg5-2.6.qwe", b"123", {"name": "Pkg5"}, code=400)
-        submit.file("pkg5-2.7.tgz", b"123", {"name": "pkg5"}, code=403)
+        r = submit.file("pkg5-2.6.qwe", b"123", {"name": "Pkg5"}, code=400)
+        assert "not a valid" in r.status
+        r = submit.file("pkg5-2.7.tgz", b"123", {"name": "pkg5"}, code=403)
+        assert "cannot register" in r.status
 
     def test_upload_use_registered_name_issue84(self, submit, mapp):
         metadata = {"name": "pkg_hello", "version":"1.0", ":action": "submit"}
@@ -354,6 +357,14 @@ class TestSubmitValidation:
         testapp.xget(404, path1)
         testapp.xget(200, path2)
 
+    def test_upload_with_metadata(self, submit, testapp, mapp, pypistage):
+        pypistage.mock_simple("package", '<a href="/package-1.0.zip" />')
+        mapp.upload_file_pypi(
+                        "package-1.0.tar.gz", b'123',
+                        "package", "1.0",
+                        indexname=submit.stagename, register=False,
+                        code=200)
+
     def test_get_project_redirected(self, submit, mapp):
         metadata = {"name": "Pkg1", "version": "1.0", ":action": "submit",
                     "description": "hello world"}
@@ -362,6 +373,23 @@ class TestSubmitValidation:
         assert location.endswith("/Pkg1")
 
 
+def test_submit_authorization(mapp, testapp):
+    from base64 import b64encode
+    import sys
+    api = mapp.create_and_use()
+    testapp.auth = None
+    data = {':action': 'submit', "name": "Pkg1", "version": "1.0"}
+    r = testapp.post(api.index + '/', data, expect_errors=True)
+    assert r.status_code == 403
+    basic_auth = '%s:%s' % (api.user, api.password)
+    basic_auth = b"Basic " + b64encode(basic_auth.encode("ascii"))
+    if sys.version_info[0] >= 3:
+        basic_auth = basic_auth.decode("ascii")
+    headers = {'Authorization': basic_auth}
+    r = testapp.post(api.index + '/', data, headers=headers)
+    assert r.status_code == 200
+
+
 def test_push_non_existent(mapp, testapp, monkeypatch):
     # check that push from non-existent index results in 404
     req = dict(name="pkg5", version="2.6", targetindex="user2/dev")
